// powers.cpp - 'A Quest Too Far'
//
// Copyright 2010 Geoffrey White
//
// see readme.txt for more details.

#include "powers.h"
#include "creature.h"
#include "map.h"
#include "fx.h"
#include "message.h"
#include "describe.h"
#include "game.h"
#include "item.h"
#include "rng.h"
#include <assert.h>

// --- defines ---

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

char ray_chars[3][3] =
{{	'\\',	'-',	'/'},
 {	'|',	'*',	'|'},
 {	'/',	'-',	'\\'}
};

// --- do_teleport ---

void do_teleport(creature *_creature)
{
	map_tile *tile;
	map *_map;
	int x, y;

	assert(_creature != NULL);
	_map = _creature->get_map();
	assert(_map != NULL);

	// find a position to teleport to
	while (true)
	{
		x = random(0, MAP_WIDTH - 1);
		y = random(0, MAP_HEIGHT - 1);
		tile = map::get_tile(_map, x, y);
		if ((_creature->can_move(tile)) ||
			((x == _creature->get_x()) && (y == _creature->get_y()))) // (so we can't get stuck)
		{
			// teleport!
			do_teleport_to(_creature, _map, x, y);
			return;
		}
	}
}

void do_teleport_to(creature *_creature, map *_map, int x, int y)
{
	bool visible;

	assert(_creature != NULL);
	assert(_map != NULL);

	visible = (_creature->is_visible()) || (_creature->is_hero());

	// teleport
	_creature->move_to(_map, x, y);

	visible |= (_creature->is_visible());

	// report
	if (visible)
	{
		stack::push_action(display_message("Pop!"));
	}
}

// --- zap_disintegration ---

void zap_disintegration(creature *_creature, int dx, int dy)
{
	fx_tile my_fx[MAP_WIDTH][MAP_HEIGHT];
	map_tile *tile;
	map *_map;
	item *_item;
	int x, y, dist;
	bool visible;

	assert(_creature != NULL);
	_map = _creature->get_map();
	assert(_map != NULL);

	// determine visibility
	visible = (_creature->is_visible()) || (_creature->is_hero());
	for (dist = 10; dist >= 1; dist--)
	{
		x = _creature->get_x() + (dist * dx);
		y = _creature->get_y() + (dist * dy);
		tile = map::get_tile(_map, x, y);
		if ((tile != NULL) && (tile->vis == V_VISIBLE))
		{
			visible = true;
		}
	}

	// post zap: wait for message, clear FX
	stack::push_action(clear_fx(*_map));
	if (visible)
	{
		stack::push_action(wait_if_message(true));
	}

	// prep fx
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			my_fx[x][y].is_fx = false;
		}
	}

	// disintegrate tiles
	for (dist = 10; dist >= 1; dist--)
	{
		x = _creature->get_x() + (dist * dx);
		y = _creature->get_y() + (dist * dy);
		tile = map::get_tile(_map, x, y);
		if (tile != NULL)
		{
			if (tile->my_creature != NULL)
			{
				// disintegrate creature
				stack::push_action(creature_damage(_creature, tile->my_creature, DT_DISINTEGRATE, 999));
			}
			while (!tile->my_items.empty())
			{
				// disintegrate item
				_item = tile->my_items.front();
				if (tile->vis == V_VISIBLE)
				{
					stack::push_action(display_message("A " + _item->short_name + " is disintegrated!"));
				}
				delete _item;
				tile->my_items.pop_front();
			}
			if (tile->is_solid)
			{
				// disintegrate wall
				tile->display_ch = ':';
				tile->is_solid = false;
			}

			// prep fx
			my_fx[x][y].is_fx = true;
			my_fx[x][y].ch = '*';
			my_fx[x][y].style = mk_char_style(255, 255, 0, 0, 0, 0);
		}
	}

	// show / describe fx
	if (visible)
	{
		stack::push_action(show_fx(*_map, &(my_fx[0][0])));
		stack::push_action(display_message("Zap!"));
	}
}

// --- zap_displacement ---

void zap_displacement(creature *_creature, int dx, int dy)
{
	map_tile *tile;
	map *_map;
	int x, y, xd, yd, dist;
	bool do_displace;

	assert(_creature != NULL);
	_map = _creature->get_map();
	assert(_map != NULL);

	// find a position to displace to
	do_displace = false;
	dist = 0;
	while (true)
	{
		dist++;

		x = _creature->get_x() + (dist * dx);
		y = _creature->get_y() + (dist * dy);
		tile = map::get_tile(_map, x, y);
		if ((tile == NULL) || (tile->is_solid))
		{
			// stop at walls
			break;
		}
		if (_creature->can_move(tile))
		{
			// we could move here
			xd = x;
			yd = y;
			do_displace = true;
		}
	}

	// displace
	if (do_displace)
	{
		do_teleport_to(_creature, _map, xd, yd);
	} else {
		if (_creature->is_hero())
		{
			stack::push_action(display_message("Nothing happens."));
		}
	}
}

// --- zap_fire ---

char_style fire_colour(int dist)
{
	int r, g, b, dist2;

	// adjust distance
	dist2 = MAX(dist - 1, 0);

	// determine colour
	r = 320;
	g = 320 - (dist2 * 64);
	b = 192 - (dist2 * 128);

	// clamp and return
	r = MAX(MIN(r, 255), 0);
	g = MAX(MIN(g, 255), 0);
	b = MAX(MIN(b, 255), 0);
	return mk_char_style(r, g, b, 0, 0, 0);
}

void zap_fire(creature *_creature, int dx, int dy)
{
	fx_tile my_fx[MAP_WIDTH][MAP_HEIGHT];
	map_tile *tile;
	map *_map;
	int x, y, i, dist, damage_amount;
	bool visible;

	assert(_creature != NULL);
	_map = _creature->get_map();
	assert(_map != NULL);

	// find the target tile and determine visibility
	dist = 0;
	visible = (_creature->is_visible()) || (_creature->is_hero());
	while (true)
	{
		dist++;

		x = _creature->get_x() + (dist * dx);
		y = _creature->get_y() + (dist * dy);
		tile = map::get_tile(_map, x, y);
		if ((tile == NULL) || (tile->is_solid))
		{
			// stop before walls
			dist--;
			break;
		}
		if (tile->vis == V_VISIBLE)
		{
			visible = true;
		}
		if (tile->my_creature != NULL)
		{
			// stop at creatures
			break;
		}
	}

	// post zap: wait for message, clear FX
	stack::push_action(clear_fx(*_map));
	if (visible)
	{
		stack::push_action(wait_if_message(true));
	}

	// prep fx
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			my_fx[x][y].is_fx = false;
		}
	}

	// do flame
	for (i = (dist == 0) ? (0) : (1); i <= dist; i++)
	{
		x = _creature->get_x() + (i * dx);
		y = _creature->get_y() + (i * dy);
		tile = map::get_tile(_map, x, y);
		if (tile != NULL)
		{
			if (tile->my_creature != NULL)
			{
				// check resistances
				if (tile->my_creature->resists_fire)
				{
					damage_amount = random(5, 10);
				} else {
					damage_amount = random(25, 30);
				}
 
				// burn creature
				stack::push_action(creature_damage(_creature, tile->my_creature, DT_FIRE, damage_amount));

				// report
				if ((tile->vis == V_VISIBLE) || (tile->my_creature->is_hero()))
				{
					stack::push_action(display_message(describe_burn(_creature, tile->my_creature, damage_amount)));
				}
			}

			// prep fx
			my_fx[x][y].is_fx = true;
			if (i == dist)
			{
				my_fx[x][y].ch = '*';
			} else {
				my_fx[x][y].ch = ray_chars[dx + 1][dy + 1];
			}
			my_fx[x][y].style = fire_colour(i);
		}
	}

	// show / describe fx
	if (visible)
	{
		stack::push_action(show_fx(*_map, &(my_fx[0][0])));
		stack::push_action(display_message("Whoosh!"));
	}
}

// --- zap_spines ---

void zap_spines(creature *_creature, int dx, int dy)
{
	fx_tile my_fx[MAP_WIDTH][MAP_HEIGHT];
	map_tile *tile;
	map *_map;
	int x, y, i, dist, damage_amount;
	bool visible;

	assert(_creature != NULL);
	_map = _creature->get_map();
	assert(_map != NULL);

	// find the target tile and determine visibility
	dist = 0;
	visible = (_creature->is_visible()) || (_creature->is_hero());
	while (true)
	{
		dist++;

		x = _creature->get_x() + (dist * dx);
		y = _creature->get_y() + (dist * dy);
		tile = map::get_tile(_map, x, y);
		if ((tile == NULL) || (tile->is_solid))
		{
			// stop before walls
			dist--;
			break;
		}
		if (tile->vis == V_VISIBLE)
		{
			visible = true;
		}
		if (tile->my_creature != NULL)
		{
			// stop at creatures
			break;
		}
	}

	// post zap: wait for message, clear FX
	stack::push_action(clear_fx(*_map));
	if (visible)
	{
		stack::push_action(wait_if_message(true));
	}

	// prep fx
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			my_fx[x][y].is_fx = false;
		}
	}

	// do spikes
	for (i = (dist == 0) ? (0) : (1); i <= dist; i++)
	{
		x = _creature->get_x() + (i * dx);
		y = _creature->get_y() + (i * dy);
		tile = map::get_tile(_map, x, y);
		if (tile != NULL)
		{
			if (tile->my_creature != NULL)
			{
				// check resistances
				damage_amount = random(4, 10);
 
				// damage creature
				stack::push_action(creature_damage(_creature, tile->my_creature, DT_NORMAL, damage_amount));

				// report
				if ((tile->vis == V_VISIBLE) || (tile->my_creature->is_hero()))
				{
					stack::push_action(display_message(describe_spine(_creature, tile->my_creature, damage_amount)));
				}
			}

			// prep fx
			if (i % 2 == 1)
			{
				my_fx[x][y].is_fx = true;
				my_fx[x][y].ch = ray_chars[dx + 1][dy + 1];
				my_fx[x][y].style = mk_char_style(192, 192, 192, 0, 0, 0);
			}
		}
	}

	// show / describe fx
	if (visible)
	{
		stack::push_action(show_fx(*_map, &(my_fx[0][0])));
		stack::push_action(display_message(describe_spines(_creature)));
	}
}

// --- do_summons ---

void do_summons(game &_game, creature *_creature)
{
	map_tile *tile;
	map *_map;
	int x, y, summoned, possible, radius;
	bool visible;

	assert(_creature != NULL);
	_map = _creature->get_map();
	assert(_map != NULL);

	visible = (_creature->is_visible()) || (_creature->is_hero());

	// find a radius at which we can summon four monsters
	radius = 0;
	possible = 0;
	while ((possible < 3) && (radius < 3))
	{
		radius++;
		possible++;
		for (y = _creature->get_y() - radius; y <= _creature->get_y() + radius; y++)
		{
			for (x = _creature->get_x() - radius; x <= _creature->get_x() + radius; x++)
			{
				tile = map::get_tile(_map, x, y);
				if ((tile != NULL) && (!tile->is_solid) && (tile->my_creature == NULL))
				{
					possible++;
				}
			}
		}
	}

	// try to summon four monsters at the determined radius
	summoned = 0;
	for (y = _creature->get_y() - radius; y <= _creature->get_y() + radius; y++)
	{
		for (x = _creature->get_x() - radius; x <= _creature->get_x() + radius; x++)
		{
			tile = map::get_tile(_map, x, y);
			if ((tile != NULL) && (!tile->is_solid) && (tile->my_creature == NULL))
			{
				if (random(possible) < 4 - summoned)
				{
					// summon a monster
					stack::push_action(summon_creature(_game, _map, x, y, CT_SPIDER, false));
					summoned++;
					visible |= (tile->vis == V_VISIBLE);
				}
				possible--;
			}
		}
	}

	// report
	if ((visible) && (summoned > 0))
	{
		if (summoned == 1)
		{
			stack::push_action(display_message("There's a puff of smoke.  A spider appears!"));
		} else {
			stack::push_action(display_message("There's a puff of smoke.  Spiders appear!"));
		}
	}
}

// --- summon_creature ---

void summon_creature :: carry_out(const char_key &key)
{
	assert(map::get_tile(my_map, x, y) != NULL);
	assert(map::get_tile(my_map, x, y)->my_creature == NULL);

	new creature(my_game, my_map, x, y, type, can_be_equipped);
}
